{$inline on}
program binterp;
uses
  sysutils, strutils;

const
  RawTagDelims = ['<', '|', '>'];

type
  TRGBQuad = packed record
    B: byte;
    G: byte;
    R: byte;
    O: byte; // Always 0
  end;
  TBMPHeader = packed record
    // BITMAPFILEHEADER
    ID: packed array[1..2] of char; // 'BM'
    FileSize: longword; // DataSize + Offset
    Res1: word;
    Res2: word;
    Offset: longword; // 1078 = SizeOf(TBMPHeader) + SizeOf(Palette)
    // BITMAPINFO <CORE>
    IHSize: longword; // 40
    Width: longword;
    Height: longword;
    Planes: word; // 1
    BitCount: word; // 8
    // </CORE>
    Compress: longword;
    DataSize: longword; // Width * Height
    XPxPerMeter: longword; // dpi * 39.37ish
    YPxPerMeter: longword; // dpi * 39.37ish
    Colors: longword; // 256
    IColors: longword; // 256
  end;
  TBGR = packed record
    B: byte;
    G: byte;
    R: byte;
  end;
  TDummyHeader = packed record
   FrameWidth: longword;
   FrameHeight: longword;
  end;
  TImagePointData = record
   X: longword;
   Y: longword;
   Color: byte;
  end;
  EInterpolationMethod = (
    imDefault,
    imNearestNeighbor
  );
  TByteBits = bitpacked array[0..7] of boolean;
  TByteNybs = bitpacked array[0..1] of 0..15;
  TUInt24 = 0..$ffffff;

//****************************************************************************//
//****************************** Misc Functions ******************************//
//****************************************************************************//

function SquaredEuclideanDistance2D (
  X1: longint; Y1: longint;
  X2: longint; Y2: longint
) : longword; inline;
begin
  SquaredEuclideanDistance2D := sqr(abs(X1 - X2)) + sqr(abs(Y1 - Y2));
end;

function SquaredEuclideanDistance3D (
  X1: longint; Y1: longint; Z1: longint;
  X2: longint; Y2: longint; Z2: longint
) : longword;
begin
  SquaredEuclideanDistance3D := sqr(abs(X1 - X2)) + sqr(abs(Y1 - Y2)) + sqr(abs(Z1 - Z2));
end;

//********************************************************//
function FastPercievedColorDistance (
  R1: longint; G1: longint; B1: longint;
  R2: longint; G2: longint; B2: longint
) : longword;
begin
  FastPercievedColorDistance := sqr(30 * abs(R1 - R2)) + sqr(59 * abs(G1 - G2)) + sqr(11 * abs(B1 - B2));
end;


//********************************************************//
function DistWt (
  X1: longint; Y1: longint;
  X2: longint; Y2: longint
) : extended; inline;
begin
  if (X1 = X2) and (Y1 = Y2) then // don't divide by zero
    DistWt := 1
  else
    DistWt := sqr(1 / ( sqr(abs(X1 - X2)) + sqr(abs(Y1 - Y2)) ));
end;


//****************************************************************************//
//*********************************** Main ***********************************//
//****************************************************************************//
var
  Header: TDummyHeader;
  PImage, PMult, Pbuf: ^byte;
  PMask: ^ByteBool;
  RawFile, PALFile: text;
  RawLine, RawTag, RawToken: AnsiString;
  BMPPalette, InPalette: array[0..255] of TRGBQuad;
  CLUT: array[0..255] of byte;
  ColorDist, ClosestColorDist: longword;
  BMPFN, InputFN, OutputFN: string;
  InputFile, OutputFile, BMPFile: file;
  BMPHeader: TBMPHeader;
  HaveMask: boolean = false;
  RemapInput: boolean = false;

  GoodPoints: array of TImagePointData;
  InterpMethod: EInterpolationMethod = imDefault;
  i, j: word;
  X, Y, Z, N: longword;
  NewColor, TotalWeight, W: extended;

begin
  if (ParamCount < 1) then
   begin
    writeln('Usage: ' + ParamStr(0) + ' <Input File>');
    halt(1)
   end
  else if not FileExists(ParamStr(1)) then
   begin
    writeln(ParamStr(1), ': File not found.');
    halt(1)
   end;
  assign(RawFile, ParamStr(1));
  reset(RawFile);

  while not eof(RawFile) do
   begin
    readln(RawFile, RawLine);
    RawLine := Trim(RawLine);

    while (pos('<', RawLine) > 0) do
     begin
      X := pos('<', RawLine);
      RawTag := copy(RawLine, X, Length(RawLine));
      SetLength(RawTag, pos('>', RawTag));
      Y := X + Length(RawTag);
      RawLine := copy(RawLine, Y, Length(RawLine));

      if (WordCount(RawTag, RawTagDelims) > 0) then
       begin
        writeln(RawTag);
        RawToken := upcase(ExtractWord(1, RawTag, RawTagDelims));

	//************************************//
        if (WordCount(RawTag, RawTagDelims) > 1) then
         begin
	//************************************//
          if (RawToken = 'METHOD') then
           begin 
            if (upcase(ExtractWord(2, RawTag, RawTagDelims)) = 'NEAREST') then
              InterpMethod := imNearestNeighbor
            else
              InterpMethod := imDefault;
           end

	//************************************//
          else if (RawToken = 'MASK') then
           begin // load mask data
            Filemode := 0; // should be Read - default = ReadWrite = run-time error when opening read-only files
            BMPFN := ExtractWord(2, RawTag, RawTagDelims);
            if FileExists(BMPFN) then
             begin
              assign(BMPFile, BMPFN);
              reset(BMPFile, 1);

              BlockRead(BMPFile, BMPHeader, SizeOf(BMPHeader));

              if (BMPHeader.ID <> 'BM') then
                writeln(BMPFN, ' is not a valid BMP file!')

              else if (BMPHeader.Compress <> 0) then
               begin
                writeln(BMPFN, ' is compressed!');
                writeln('Could not load image data.');
               end

              else
               begin
                if HaveMask then
                 begin
                  FreeMem(PMask, (Header.FrameWidth * Header.FrameHeight));
                  FreeMem(PImage, (Header.FrameWidth * Header.FrameHeight));
                  FreeMem(PMult, (Header.FrameWidth * Header.FrameHeight));
                  FreeMem(Pbuf, (Header.FrameWidth * 4));
                 end;

                HaveMask := true;

                Header.FrameWidth := BMPHeader.Width;
                Header.FrameHeight := BMPHeader.Height;

                PMask := GetMem((Header.FrameWidth * Header.FrameHeight));
                PImage := GetMem((Header.FrameWidth * Header.FrameHeight));
                PMult := GetMem((Header.FrameWidth * Header.FrameHeight));
                FillByte(PMult^, (Header.FrameWidth * Header.FrameHeight), $ff);
                Pbuf := GetMem((Header.FrameWidth * 4));

                seek(BMPFile, BMPHeader.Offset);

                if (BMPHeader.BitCount = 1) then
                  for Y := (Header.FrameHeight - 1) downto 0 do
                   begin
                    Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                    writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                    BlockRead(BMPFile, Pbuf^, (BMPHeader.DataSize DIV BMPHeader.Height));
                    for X := 0 to (Header.FrameWidth - 1) do
                      PMask[(Z + X)] := TByteBits(Pbuf[(X DIV 8)])[((not X) and 7)];
                   end

                else if (BMPHeader.BitCount = 4) then
                  for Y := (Header.FrameHeight - 1) downto 0 do
                   begin
                    Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                    writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                    BlockRead(BMPFile, Pbuf^, (BMPHeader.DataSize DIV BMPHeader.Height));
                    for X := 0 to (Header.FrameWidth - 1) do
                      PMask[(Z + X)] := (TByteNybs(Pbuf[(X DIV 2)])[((not X) and 1)] > 0);
                   end

                else if (BMPHeader.BitCount = 8) then
                  for Y := (Header.FrameHeight - 1) downto 0 do
                   begin
                    Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                    writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                    BlockRead(BMPFile, Pbuf^, (BMPHeader.DataSize DIV BMPHeader.Height));
                    Move(Pbuf^, PMask[Z], Header.FrameWidth);
                   end

                else if (BMPHeader.BitCount = 24) then
                  for Y := (Header.FrameHeight - 1) downto 0 do
                   begin
                    Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                    writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                    BlockRead(BMPFile, Pbuf^, (BMPHeader.DataSize DIV BMPHeader.Height));
                    for X := 0 to (Header.FrameWidth - 1) do
                      PMask[(Z + X)] := ByteBool(Pbuf[(X * 3)] or Pbuf[(X * 3 + 1)] or Pbuf[(X * 3 + 2)]);
                   end

                else
                 begin
                  writeln(BMPFN, ' has an unsupported color depth.');
                  writeln('The recommended color depth is 256 colors/8 bits per pixel.');
                  HaveMask := false;
                 end;
               end;

              close(BMPFile);
             end
            else
             begin
              writeln(BMPFN, ': File not found.');
              HaveMask := false;
             end;

            //**************************//
            if HaveMask and (WordCount(RawTag, RawTagDelims) > 2) then
             begin
              BMPFN := ExtractWord(3, RawTag, RawTagDelims);
              if FileExists(BMPFN) then
               begin
                assign(BMPFile, BMPFN);
                reset(BMPFile, 1);

                BlockRead(BMPFile, BMPHeader, SizeOf(BMPHeader));

                if (BMPHeader.ID <> 'BM') then
                  writeln(BMPFN, ' is not a valid BMP file!')

                else if (BMPHeader.Compress <> 0) then
                 begin
                  writeln(BMPFN, ' is compressed!');
                  writeln('Could not load image data.');
                 end

                else
                 begin
                  if (BMPHeader.IHSize > 40) then // fucking ImageMagick
                    seek(BMPFile, (FilePos(BMPFile) + BMPHeader.IHSize - 40));

                  if (BMPHeader.BitCount = 1) then
                   begin
                    BlockRead(BMPFile, BMPPalette, 8);
                    for i := 0 to 1 do
                      CLUT[i] := ((BMPPalette[i].R * 30) + (BMPPalette[i].G * 59) + (BMPPalette[i].B * 11)) DIV 100;

                    seek(BMPFile, BMPHeader.Offset);
                    for Y := (Header.FrameHeight - 1) downto 0 do
                     begin
                      Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                      writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                      BlockRead(BMPFile, Pbuf^, (BMPHeader.DataSize DIV BMPHeader.Height));
                      for X := 0 to (Header.FrameWidth - 1) do
                        PMult[(Z + X)] := CLUT[ord(TByteBits(Pbuf[(X DIV 8)])[((not X) and 7)])];
                     end;
                   end

                  else if (BMPHeader.BitCount = 4) then
                   begin
                    BlockRead(BMPFile, BMPPalette, 64);
                    for i := 0 to 15 do
                      CLUT[i] := ((BMPPalette[i].R * 30) + (BMPPalette[i].G * 59) + (BMPPalette[i].B * 11)) DIV 100;

                    seek(BMPFile, BMPHeader.Offset);

                    for Y := (Header.FrameHeight - 1) downto 0 do
                     begin
                      Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                      writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                      BlockRead(BMPFile, Pbuf^, (BMPHeader.DataSize DIV BMPHeader.Height));
                      for X := 0 to (Header.FrameWidth - 1) do
                        PMult[(Z + X)] := CLUT[TByteNybs(Pbuf[(X DIV 2)])[((not X) and 1)]];
                     end;
                   end

                  else if (BMPHeader.BitCount = 8) then
                   begin
                    BlockRead(BMPFile, BMPPalette, SizeOf(BMPPalette));
                    for i := 0 to 255 do
                      CLUT[i] := ((BMPPalette[i].R * 30) + (BMPPalette[i].G * 59) + (BMPPalette[i].B * 11)) DIV 100;

                    seek(BMPFile, BMPHeader.Offset);
                    for Y := (Header.FrameHeight - 1) downto 0 do
                     begin
                      Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                      writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                      BlockRead(BMPFile, Pbuf^, (BMPHeader.DataSize DIV BMPHeader.Height));
                      for X := 0 to (Header.FrameWidth - 1) do
                        PMult[Z + X] := CLUT[Pbuf[X]];
                     end;
                   end

                  else if (BMPHeader.BitCount = 24) then
                   begin
                    seek(BMPFile, BMPHeader.Offset);
                    for Y := (Header.FrameHeight - 1) downto 0 do
                     begin
                      Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                      writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                      BlockRead(BMPFile, Pbuf^, (BMPHeader.DataSize DIV BMPHeader.Height));
                      for X := 0 to (Header.FrameWidth - 1) do
                        PMult[(Z + X)] := ((Pbuf[(X * 3)] * 11) + (Pbuf[(X * 3 + 1)] * 59) + (Pbuf[(X * 3 + 2)] * 30)) DIV 100;
                     end;
                   end

                  else
                   begin
                    writeln(BMPFN, ' has an unsupported color depth.');
                    writeln('The recommended color depth is 256 colors/8 bits per pixel.');
                   end;
                 end;

                close(BMPFile);
               end
              else
                writeln(BMPFN, ': File not found.');

             end;

            Filemode := 2; // back to ReadWrite - just in case leaving it as Read does something even stupider
           end

	//************************************//
          else if (RawToken = 'PALETTE_FROM_BMP') then
           begin // load palette from a BMP file
            Filemode := 0; // Read
            RemapInput := false;

            BMPFN := ExtractWord(2, RawTag, RawTagDelims);
            if FileExists(BMPFN) then
             begin
              assign(BMPFile, BMPFN);
              reset(BMPFile, 1);

              BlockRead(BMPFile, BMPHeader, SizeOf(BMPHeader));

              if (BMPHeader.ID <> 'BM') then
                writeln(BMPFN, ' is not a valid BMP file!')

              else if (BMPHeader.BitCount = 8) then
               begin
                if (BMPHeader.IHSize > 40) then // fucking ImageMagick
                  seek(BMPFile, (FilePos(BMPFile) + BMPHeader.IHSize - 40));

                BlockRead(BMPFile, InPalette, SizeOf(InPalette));
                RemapInput := true;
               end

              else if (BMPHeader.BitCount = 24) and ((BMPHeader.Width * BMPHeader.Height) = 256) then
                for Y := 0 to (BMPHeader.Height - 1) do
                 begin
                  seek(BMPFile, (Y * (BMPHeader.DataSize DIV BMPHeader.Height) + BMPHeader.Offset));
                  for X := 0 to (BMPHeader.Width - 1) do
                    BlockRead(BMPFile, InPalette[((Y * BMPHeader.Width) + X)], 3);
                  RemapInput := true;
                 end

              else
                writeln(BMPFN, ' does not have a 256-color palette.');

              close(BMPFile);
             end
            else
              writeln(BMPFN, ': File not found.');

            Filemode := 2; // ReadWrite
           end
{
	//************************************//
          else if (RawToken = 'PALETTE_JASC') then
           begin // load a JASC palette
            Filemode := 0; // Read
            RemapInput := false;

            InputFN := ExtractWord(2, RawTag, RawTagDelims);
            if FileExists(InputFN) then
             begin 
              assign(PALFile, InputFN);
              reset(PALFile);


              close(PALFile);
              RemapInput := true;
             end;

            Filemode := 2; // ReadWrite
           end
}
	//************************************//
          else if HaveMask and (RawToken = 'IMAGE') then
           begin // load overlay data

            SetLength(GoodPoints, 0);
            BMPFN := ExtractWord(2, RawTag, RawTagDelims);
            if FileExists(BMPFN) then
             begin
              Filemode := 0; // Read
              assign(BMPFile, BMPFN);
              reset(BMPFile, 1);
              Filemode := 2; // ReadWrite

              BlockRead(BMPFile, BMPHeader, SizeOf(BMPHeader));

              if (BMPHeader.ID <> 'BM') then
                writeln(BMPFN, ' is not a valid BMP file!')

              else if (BMPHeader.Width <> Header.FrameWidth) or (BMPHeader.Height <> Header.FrameHeight) then
               begin
                writeln(BMPFN, ' has the wrong dimensions!');
                writeln('Correct dimensions for current mask are: ', Header.FrameWidth, 'x', Header.FrameHeight);
               end

              else if (BMPHeader.Compress <> 0) then
               begin
                writeln(BMPFN, ' is compressed!');
                writeln('Could not load image data.');
               end

              else if (BMPHeader.BitCount = 8) then
               begin
                if (BMPHeader.IHSize > 40) then // fucking ImageMagick
                  seek(BMPFile, (FilePos(BMPFile) + BMPHeader.IHSize - 40));

                BlockRead(BMPFile, BMPPalette, SizeOf(BMPPalette));
                if RemapInput then
                  for i := 0 to 255 do
                   begin // populate Color Look-Up Table
                    ClosestColorDist := $ffffffff;
                    for j := 0 to 255 do
                     begin
                      ColorDist := FastPercievedColorDistance(BMPPalette[i].R, BMPPalette[i].G, BMPPalette[i].B, InPalette[j].R, InPalette[j].G, InPalette[j].B);
                      if (ColorDist < ClosestColorDist) then
                       begin
                        ClosestColorDist := ColorDist;
                        CLUT[i] := j;
                       end;
                      if (ClosestColorDist = 0) then
                        break;
                     end;
                    //writeln('Mapped BMP color ', i, ' (#', HexStr(BMPPalette[i].R, 2), HexStr(BMPPalette[i].G, 2), HexStr(BMPPalette[i].B, 2), ') to VID color ', CLUT[i], ' (#', HexStr((Palette[CLUT[i]].R * 4), 2), HexStr((Palette[CLUT[i]].G * 4), 2), HexStr((Palette[CLUT[i]].B * 4), 2), ').');
                   end
                else
                  for i := 0 to 255 do
                    CLUT[i] := i;

                seek(BMPFile, BMPHeader.Offset);
                for Y := (Header.FrameHeight - 1) downto 0 do
                 begin
                  Z := (Y * Header.FrameWidth);
{$IFDEF BUGHUNT}
                  writeln('Row ', Y, ': ', FilePos(BMPFile), ' -> ', Z);
{$ENDIF}
                  BlockRead(BMPFile, Pbuf^, (BMPHeader.DataSize DIV BMPHeader.Height));
                  //Move(Pbuf^, PImage[Z], Header.FrameWidth);
                  for X := 0 to (Header.FrameWidth - 1) do
                    if PMask[(Z + X)] then
                     begin
                      PImage[(Z + X)] := CLUT[Pbuf[X]];
                      SetLength(GoodPoints, (Length(GoodPoints) + 1));
                      GoodPoints[High(GoodPoints)].X := X;
                      GoodPoints[High(GoodPoints)].Y := Y;
                      GoodPoints[High(GoodPoints)].Color := CLUT[Pbuf[X]];
                     end;
                 end;

                for Y := 0 to (Header.FrameHeight - 1) do
                 begin
                  Z := (Y * Header.FrameWidth);

                  for X := 0 to (Header.FrameWidth - 1) do
                    if (PMult[(Z + X)] > 0) then
                     begin
                      if not PMask[(Z + X)] then
                       begin
                        if (InterpMethod = imNearestNeighbor) then
                         begin
                          ClosestColorDist := $ffffffff;
                          for N := Low(GoodPoints) to High(GoodPoints) do
                           begin
                            ColorDist := SquaredEuclideanDistance2D(X, Y, GoodPoints[N].X, GoodPoints[N].Y);
                            if (ColorDist < ClosestColorDist) then
                             begin
                              ClosestColorDist := ColorDist;
                              PImage[(Z + X)] := round(((PMult[(Z + X)] / 255) * GoodPoints[N].Color));
                             end;
                           end;
                         end
                        else // default interpolation
                         begin
                          NewColor := 0;
                          TotalWeight := 0;

                          for N := Low(GoodPoints) to High(GoodPoints) do
                           begin
                            W := DistWt(X, Y, GoodPoints[N].X, GoodPoints[N].Y);
                            TotalWeight := TotalWeight + W;
                            NewColor := NewColor + (W * GoodPoints[N].Color);
                           end;

                          NewColor := NewColor * (1 / TotalWeight);
                          PImage[(Z + X)] := round(((PMult[(Z + X)] / 255) * NewColor));
                         end;
                       end
                      else
                       PImage[(Z + X)] := round(((PMult[(Z + X)] / 255) * PImage[(Z + X)]));
                     end
                    else
                      PImage[(Z + X)] := 0;
                 end;

                if (WordCount(RawTag, RawTagDelims) > 2) then
                  OutputFN := ExtractWord(3, RawTag, RawTagDelims)
                else
                  OutputFN := BMPFN + '_funterped.bmp';

                assign(OutputFile, OutputFN);
                rewrite(OutputFile, 1);

                writeln(' Writing ', OutputFN);

                BMPHeader.Offset := SizeOf(BMPHeader) + SizeOf(BMPPalette);
                BMPHeader.FileSize :=  BMPHeader.DataSize + BMPHeader.Offset;
                BMPHeader.IHSize := 40;

                BlockWrite(OutputFile, BMPHeader, SizeOf(BMPHeader));
                if RemapInput then
                  BlockWrite(OutputFile, InPalette, SizeOf(InPalette))
                else
                  BlockWrite(OutputFile, BMPPalette, SizeOf(BMPPalette));

                for Y := (Header.FrameHeight - 1) downto 0 do
                 begin
                  Z := (Y * Header.FrameWidth);
                  Move(PImage[Z], Pbuf^, Header.FrameWidth);
                  BlockWrite(OutputFile, Pbuf^, (BMPHeader.DataSize DIV BMPHeader.Height));
                 end;

                close(OutputFile);

               end // 8-bit BMP

              else
               begin
                writeln(BMPFN, ' has an unsupported color depth.');
                writeln('The recommended color depth is 256 colors/8 bits per pixel.');
               end;


              close(BMPFile);
             end
            else
             writeln(BMPFN, ': File not found.');

            SetLength(GoodPoints, 0);
           end;

	//************************************//
         end; // tags with at least one parameter

       end; // non-empty tags

     end; // tags in line

   end; // lines in file

  if HaveMask then
   begin
    FreeMem(PMask, (Header.FrameWidth * Header.FrameHeight));
    FreeMem(PImage, (Header.FrameWidth * Header.FrameHeight));
    FreeMem(PMult, (Header.FrameWidth * Header.FrameHeight));
    FreeMem(Pbuf, (Header.FrameWidth * 4));
   end;

  close(RawFile);

end.
